#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gsl/gsl_rng.h>
#include <cairo/cairo.h>
#include <cairo/cairo-pdf.h>

#define A4

#define pi 3.141592653589793
#define count 200
#define factor 3.0

#ifdef A4

#define pt 1
#define inch (72 * pt)
#define mm (inch / 25.4)
#define sizex (7 * inch)
#define sizey (10 * inch)
#define pagex (210 * mm)
#define pagey (297 * mm)

#else
#ifdef POSTCARD

#define inch 300
#define pt (inch / 72.0)
#define sizex 1287
#define sizey 1795
#define pagex sizex
#define pagey sizey

#else
#ifdef VIDEO

#define inch 100
#define pt (inch / 72.0)
#define sizex 1920
#define sizey 1080
#define pagex sizex
#define pagey sizey
#define offsetx (pagex / 2.0)
#define offsety (pagey / 2.0)
#undef count
#define count 1300
#define density 200
#define frames (25 * 20)
#define tframes 25
#undef factor
#define factor 4.0

#endif
#endif
#endif

#define scale (sizex / factor)
#ifndef offsetx
#define offsetx (pagex / 2.0)
#endif
#ifndef offsety
#define offsety (pagex / 2.0)
#endif
#define clipx ((pagex - sizex) / 2.0)
#define clipy clipx

gsl_rng *rng = 0;

double clamp(double x) { return fmin(fmax(x, 0), 1); }
double coordx(double x) { return scale * x + offsetx; }
double coordy(double y) { return scale * y + offsety; }
double dist(double d) { return scale * d; }

struct vec2 { double v[2]; };
struct vec3 { double v[3]; };
struct mat33 { struct vec3 v[3]; };

double vec3_normSq(struct vec3 *v) {
  return v->v[0] * v->v[0] + v->v[1] * v->v[1] + v->v[2] * v->v[2];
}

double vec2_normSq(struct vec2 *v) {
  return v->v[0] * v->v[0] + v->v[1] * v->v[1];
}

double vec3_norm(struct vec3 *v) {
  return sqrt(vec3_normSq(v));
}

void vec2_copy(struct vec2 *v, const struct vec2 *x) {
  v->v[0] = x->v[0];
  v->v[1] = x->v[1];
}

void vec3_copy(struct vec3 *v, const struct vec3 *x) {
  v->v[0] = x->v[0];
  v->v[1] = x->v[1];
  v->v[2] = x->v[2];
}

void mat33_copy(struct mat33 *v, const struct mat33 *x) {
  vec3_copy(&v->v[0], &x->v[0]);
  vec3_copy(&v->v[1], &x->v[1]);
  vec3_copy(&v->v[2], &x->v[2]);
}

void vec3_div(struct vec3 *v, double r) {
  v->v[0] /= r;
  v->v[1] /= r;
  v->v[2] /= r;
}

void vec3_mul(struct vec3 *v, double r) {
  v->v[0] *= r;
  v->v[1] *= r;
  v->v[2] *= r;
}

void vec2_mul(struct vec2 *v, double r) {
  v->v[0] *= r;
  v->v[1] *= r;
}

void vec3_add(struct vec3 *v, const struct vec3 *l, const struct vec3 *r) {
  v->v[0] = l->v[0] + r->v[0];
  v->v[1] = l->v[1] + r->v[1];
  v->v[2] = l->v[2] + r->v[2];
}

void vec3_sub(struct vec3 *v, const struct vec3 *l, const struct vec3 *r) {
  v->v[0] = l->v[0] - r->v[0];
  v->v[1] = l->v[1] - r->v[1];
  v->v[2] = l->v[2] - r->v[2];
}
void vec2_sub(struct vec2 *v, const struct vec2 *l, const struct vec2 *r) {
  v->v[0] = l->v[0] - r->v[0];
  v->v[1] = l->v[1] - r->v[1];
}

void vec3_normalize(struct vec3 *v) {
  double r = vec3_norm(v);
  if (1.0/0.0 > r && r > 0) {
    vec3_div(v, r);
  } else {
    v->v[0] = 0;
    v->v[1] = 0;
    v->v[2] = 0;
  }
}

void vec3_cross(struct vec3 *v, const struct vec3 *l, const struct vec3 *r) {
  v->v[0] = l->v[1] * r->v[2] - l->v[2] * r->v[1];
  v->v[1] = l->v[2] * r->v[0] - l->v[0] * r->v[2];
  v->v[2] = l->v[0] * r->v[1] - l->v[1] * r->v[0];
}

double vec2_cross(const struct vec2 *u, const struct vec2 *v) {
  return u->v[0] * v->v[1] - u->v[1] * v->v[0];
}

double double_random() {
  return gsl_rng_uniform(rng);
}

void vec3_random(struct vec3 *v) {
  v->v[0] = double_random() - 0.5;
  v->v[1] = double_random() - 0.5;
  v->v[2] = double_random() - 0.5;
  vec3_normalize(v);
}

void mat33_multmm(struct mat33 *r, const struct mat33 *x, const struct mat33 *y) {
  for (int i = 0; i < 3; ++i) {
    for (int j = 0; j < 3; ++j) {
      r->v[i].v[j] = 0;
      for (int k = 0; k < 3; ++k) {
        r->v[i].v[j] += x->v[i].v[k] * y->v[k].v[j];
      }
    }
  }
}

void mat33_multmv(struct vec3 *r, const struct mat33 *x, const struct vec3 *y) {
  for (int i = 0; i < 3; ++i) {
    r->v[i] = 0;
    for (int k = 0; k < 3; ++k) {
      r->v[i] += x->v[i].v[k] * y->v[k];
    }
  }
}

void mat33_random(struct mat33 *v, struct mat33 *v1, double k) {
  double a = k * double_random();
  double b = k * double_random();
  double c = k * double_random();
  double ca = cos(a);
  double sa = sin(a);
  double cb = cos(b);
  double sb = sin(b);
  double cc = cos(c);
  double sc = sin(c);
  struct mat33 ma = {{{{ca,sa,0}},{{-sa,ca,0}},{{0,0,1}}}};
  struct mat33 mb = {{{{cb,0,sb}},{{0,1,0}},{{-sb,0,cb}}}};
  struct mat33 mc = {{{{1,0,0}},{{0,cc,sc}},{{0,-sc,cc}}}};
  struct mat33 ma1 = {{{{ca,-sa,0}},{{sa,ca,0}},{{0,0,1}}}};
  struct mat33 mb1 = {{{{cb,0,-sb}},{{0,1,0}},{{sb,0,cb}}}};
  struct mat33 mc1 = {{{{1,0,0}},{{0,cc,-sc}},{{0,sc,cc}}}};
  struct mat33 t;
  mat33_multmm(&t, &ma, &mb);
  mat33_multmm(v, &t, &mc);
  mat33_multmm(&t, &mb1, &ma1);
  mat33_multmm(v1, &mc1, &t);
}

double mat33_det(const struct mat33 *v) {
  double a = v->v[0].v[0];
  double b = v->v[0].v[1];
  double c = v->v[0].v[2];
  double d = v->v[1].v[0];
  double e = v->v[1].v[1];
  double f = v->v[1].v[2];
  double g = v->v[2].v[0];
  double h = v->v[2].v[1];
  double i = v->v[2].v[2];
  return a*e*i-a*f*h-b*d*i+b*f*g+c*d*h-c*e*g;
}

const struct mat33 poly_5_3[60] = {{{{{1.0,1.4865237330654327e-32,-1.232595164407831e-32}},{{1.4865237330654327e-32,1.0,2.465190328815662e-32}},{{-1.232595164407831e-32,2.465190328815662e-32,1.0}}}},{{{{0.7453559924999302,0.33333333333333326,-0.5773502691896255}},{{-0.6666666666666663,0.3726779962499651,-0.6454972243679032}},{{-7.535260718433166e-17,0.8660254037844388,0.49999999999999983}}}},{{{{0.7453559924999302,-0.6666666666666663,7.535260718433166e-17}},{{-0.6666666666666663,-0.7453559924999302,1.972756867494999e-16}},{{-7.535260718433166e-17,-1.972756867494999e-16,-1.0}}}},{{{{0.7453559924999302,-0.6666666666666663,-7.535260718433166e-17}},{{0.33333333333333326,0.3726779962499651,0.8660254037844388}},{{-0.5773502691896255,-0.6454972243679032,0.49999999999999983}}}},{{{{1.0,0.0,0.0}},{{0.0,-0.5000000000000002,-0.8660254037844387}},{{0.0,0.8660254037844387,-0.5}}}},{{{{1.0,0.0,0.0}},{{0.0,-0.5000000000000002,0.8660254037844387}},{{0.0,-0.8660254037844387,-0.5}}}},{{{{0.33333333333333404,-0.12732200375003486,-0.9341723589627156}},{{-0.7453559924999298,-0.6423503277082808,-0.1784110448865457}},{{-0.5773502691896256,0.7557613140761709,-0.309016994374948}}}},{{{{0.33333333333333404,-0.7453559924999298,0.5773502691896256}},{{-0.7453559924999298,0.1666666666666661,0.6454972243679034}},{{-0.5773502691896256,-0.6454972243679032,-0.49999999999999967}}}},{{{{0.7453559924999302,0.33333333333333315,0.5773502691896255}},{{-0.6666666666666663,0.3726779962499652,0.6454972243679031}},{{-7.535260718433166e-17,-0.8660254037844387,0.5000000000000003}}}},{{{{0.3333333333333339,-0.7453559924999298,-0.5773502691896256}},{{-0.74535599249993,0.16666666666666582,-0.6454972243679031}},{{0.5773502691896254,0.6454972243679032,-0.5}}}},{{{{0.33333333333333404,-0.7453559924999298,-0.5773502691896256}},{{-0.12732200375003483,-0.6423503277082809,0.755761314076171}},{{-0.9341723589627156,-0.17841104488654566,-0.30901699437494806}}}},{{{{0.7453559924999302,0.33333333333333326,0.5773502691896255}},{{0.33333333333333326,0.5636610018750176,-0.755761314076171}},{{-0.5773502691896255,0.755761314076171,0.3090169943749479}}}},{{{{0.7453559924999302,0.33333333333333337,-0.5773502691896255}},{{0.33333333333333326,-0.9363389981249828,-0.11026408970826784}},{{-0.5773502691896255,-0.11026408970826745,-0.8090169943749478}}}},{{{{0.7453559924999302,-0.6666666666666663,-7.535260718433166e-17}},{{0.33333333333333315,0.3726779962499651,-0.8660254037844388}},{{0.5773502691896256,0.6454972243679031,0.5000000000000003}}}},{{{{0.7453559924999302,0.33333333333333326,-0.5773502691896255}},{{0.33333333333333326,0.5636610018750176,0.755761314076171}},{{0.5773502691896255,-0.755761314076171,0.3090169943749479}}}},{{{{0.3333333333333338,-0.12732200375003533,0.9341723589627156}},{{-0.12732200375003533,0.9756836610416141,0.17841104488654563}},{{-0.9341723589627156,-0.17841104488654547,0.3090169943749482}}}},{{{{0.33333333333333404,0.8726779962499648,0.35682208977309}},{{-0.7453559924999298,0.47568366104161486,-0.46708617948135767}},{{-0.5773502691896256,-0.11026408970826751,0.8090169943749478}}}},{{{{-0.3333333333333325,-0.8726779962499653,-0.3568220897730902}},{{-0.8726779962499654,0.14235032770827957,0.4670861794813575}},{{-0.35682208977309043,0.4670861794813573,-0.8090169943749477}}}},{{{{0.333333333333334,0.8726779962499649,-0.35682208977309016}},{{-0.7453559924999299,0.4756836610416148,0.4670861794813573}},{{0.5773502691896255,0.11026408970826801,0.809016994374948}}}},{{{{0.3333333333333337,-0.12732200375003533,-0.9341723589627156}},{{-0.12732200375003572,0.9756836610416141,-0.17841104488654574}},{{0.9341723589627156,0.17841104488654586,0.30901699437494795}}}},{{{{0.3333333333333339,-0.12732200375003475,0.9341723589627156}},{{-0.74535599249993,-0.6423503277082806,0.1784110448865458}},{{0.5773502691896254,-0.7557613140761711,-0.3090169943749478}}}},{{{{0.33333333333333404,0.8726779962499649,-0.35682208977309}},{{-0.12732200375003483,-0.33333333333333326,-0.934172358962716}},{{-0.9341723589627156,0.3568220897730908,-2.7755575615628914e-16}}}},{{{{0.333333333333334,-0.7453559924999298,0.5773502691896254}},{{0.8726779962499649,0.47568366104161486,0.11026408970826812}},{{-0.35682208977309016,0.4670861794813573,0.8090169943749481}}}},{{{{0.3333333333333339,-0.12732200375003475,-0.9341723589627156}},{{0.8726779962499649,-0.3333333333333336,0.3568220897730907}},{{-0.3568220897730902,-0.9341723589627158,-5.551115123125783e-16}}}},{{{{0.33333333333333404,-0.7453559924999298,-0.5773502691896256}},{{0.8726779962499648,0.47568366104161497,-0.11026408970826762}},{{0.35682208977309005,-0.46708617948135767,0.809016994374948}}}},{{{{0.7453559924999302,0.33333333333333326,0.5773502691896255}},{{0.33333333333333315,-0.9363389981249828,0.11026408970826779}},{{0.5773502691896256,0.11026408970826784,-0.8090169943749479}}}},{{{{0.33333333333333404,-0.7453559924999298,0.5773502691896256}},{{-0.12732200375003483,-0.6423503277082809,-0.755761314076171}},{{0.9341723589627156,0.17841104488654566,-0.30901699437494806}}}},{{{{-0.33333333333333265,-0.8726779962499656,0.3568220897730894}},{{0.12732200375003458,0.333333333333333,0.9341723589627162}},{{-0.9341723589627164,0.3568220897730886,-1.7740109019667067e-16}}}},{{{{-0.33333333333333237,0.7453559924999305,-0.5773502691896265}},{{-0.8726779962499654,-0.4756836610416135,-0.11026408970826873}},{{-0.3568220897730904,0.4670861794813585,0.8090169943749471}}}},{{{{-0.3333333333333328,0.1273220037500342,-0.9341723589627164}},{{-0.8726779962499657,0.333333333333333,0.35682208977308866}},{{0.35682208977308905,0.9341723589627163,-4.163336342344339e-16}}}},{{{{-0.3333333333333325,0.12732200375003536,0.9341723589627163}},{{-0.8726779962499654,0.33333333333333354,-0.35682208977308877}},{{-0.35682208977309043,-0.9341723589627157,5.551115123125785e-16}}}},{{{{-0.33333333333333226,-0.8726779962499656,0.3568220897730903}},{{-0.8726779962499656,0.14235032770827968,-0.46708617948135717}},{{0.3568220897730898,-0.4670861794813577,-0.8090169943749481}}}},{{{{0.3333333333333337,0.8726779962499651,0.35682208977308966}},{{-0.12732200375003572,-0.33333333333333287,0.9341723589627162}},{{0.9341723589627156,-0.3568220897730908,4.887327245784233e-16}}}},{{{{-0.3333333333333325,0.12732200375003536,-0.9341723589627163}},{{0.127322003750035,-0.9756836610416143,-0.17841104488654494}},{{-0.9341723589627164,-0.17841104488654475,0.3090169943749462}}}},{{{{-0.33333333333333226,-0.8726779962499656,-0.3568220897730903}},{{0.7453559924999302,-0.47568366104161347,0.46708617948135867}},{{-0.5773502691896264,-0.11026408970826877,0.8090169943749472}}}},{{{{0.333333333333334,0.8726779962499649,0.35682208977309016}},{{0.8726779962499649,-0.14235032770828115,-0.46708617948135855}},{{-0.35682208977309016,0.46708617948135855,-0.8090169943749472}}}},{{{{0.3333333333333337,-0.12732200375003533,0.9341723589627156}},{{0.8726779962499652,-0.33333333333333276,-0.3568220897730908}},{{0.35682208977308927,0.9341723589627162,4.163336342344337e-16}}}},{{{{0.33333333333333404,0.8726779962499649,-0.35682208977309}},{{0.8726779962499648,-0.14235032770828152,0.46708617948135844}},{{0.35682208977309005,-0.4670861794813583,-0.8090169943749477}}}},{{{{-0.3333333333333325,-0.8726779962499653,-0.3568220897730902}},{{0.127322003750035,0.3333333333333334,-0.9341723589627159}},{{0.9341723589627164,-0.3568220897730886,3.885780586188048e-16}}}},{{{{-0.7453559924999292,-0.3333333333333342,-0.5773502691896268}},{{-0.3333333333333338,-0.5636610018750177,0.7557613140761708}},{{-0.577350269189627,0.7557613140761706,0.30901699437494623}}}},{{{{-0.33333333333333265,0.7453559924999299,0.5773502691896266}},{{0.12732200375003458,0.642350327708281,-0.7557613140761708}},{{-0.9341723589627164,-0.17841104488654455,-0.3090169943749463}}}},{{{{-0.7453559924999291,-0.33333333333333465,0.5773502691896267}},{{-0.33333333333333454,0.9363389981249823,0.11026408970826873}},{{-0.5773502691896265,-0.1102640897082686,-0.8090169943749471}}}},{{{{-0.3333333333333328,0.7453559924999309,0.5773502691896254}},{{-0.8726779962499657,-0.47568366104161297,0.11026408970826823}},{{0.35682208977308905,-0.46708617948135794,0.8090169943749482}}}},{{{{-0.7453559924999296,-0.3333333333333348,-0.5773502691896263}},{{-0.3333333333333348,0.9363389981249818,-0.11026408970826831}},{{0.577350269189626,0.11026408970826884,-0.8090169943749476}}}},{{{{-0.33333333333333276,0.7453559924999299,-0.5773502691896266}},{{0.1273220037500342,0.642350327708281,0.755761314076171}},{{0.9341723589627164,0.17841104488654494,-0.30901699437494606}}}},{{{{-0.3333333333333328,0.1273220037500342,0.9341723589627164}},{{0.7453559924999297,0.6423503277082814,0.17841104488654477}},{{-0.5773502691896268,0.7557613140761708,-0.3090169943749461}}}},{{{{-0.33333333333333226,0.7453559924999307,-0.5773502691896261}},{{0.7453559924999302,-0.16666666666666755,-0.6454972243679029}},{{-0.5773502691896264,-0.6454972243679022,-0.5000000000000007}}}},{{{{-0.33333333333333276,-0.8726779962499656,0.3568220897730894}},{{0.745355992499931,-0.4756836610416129,-0.46708617948135783}},{{0.5773502691896254,0.1102640897082684,0.8090169943749482}}}},{{{{-0.3333333333333325,0.12732200375003536,-0.9341723589627163}},{{0.7453559924999306,0.6423503277082809,-0.17841104488654386}},{{0.5773502691896261,-0.7557613140761711,-0.3090169943749468}}}},{{{{-0.3333333333333325,0.12732200375003536,0.9341723589627163}},{{0.127322003750035,-0.9756836610416143,0.17841104488654494}},{{0.9341723589627164,0.17841104488654475,0.3090169943749462}}}},{{{{-0.7453559924999292,0.6666666666666681,-2.597209085171972e-16}},{{-0.3333333333333338,-0.3726779962499647,-0.8660254037844389}},{{-0.577350269189627,-0.6454972243679019,0.5000000000000004}}}},{{{{-1.0000000000000004,-2.035903140772674e-15,-7.949909648767016e-16}},{{-1.554312234475219e-15,0.5,0.8660254037844392}},{{-1.2965979342720042e-15,0.8660254037844387,-0.5000000000000002}}}},{{{{-0.74535599249993,0.6666666666666671,-1.591988538067385e-15}},{{-0.333333333333335,-0.37267799624996445,0.8660254037844385}},{{0.5773502691896253,0.6454972243679031,0.5000000000000011}}}},{{{{-0.7453559924999296,-0.33333333333333287,0.5773502691896274}},{{-0.3333333333333348,-0.5636610018750179,-0.7557613140761702}},{{0.577350269189626,-0.7557613140761712,0.3090169943749467}}}},{{{{-0.74535599249993,-0.33333333333333504,0.5773502691896252}},{{0.6666666666666672,-0.37267799624996434,0.6454972243679034}},{{-1.6653345369377348e-15,0.8660254037844385,0.5000000000000009}}}},{{{{-0.7453559924999296,-0.33333333333333287,-0.5773502691896274}},{{0.6666666666666677,-0.3726779962499651,-0.645497224367902}},{{-1.2212453270876722e-15,-0.8660254037844393,0.49999999999999994}}}},{{{{-0.33333333333333276,0.7453559924999299,0.5773502691896266}},{{0.745355992499931,-0.166666666666667,0.645497224367902}},{{0.5773502691896254,0.6454972243679031,-0.5000000000000009}}}},{{{{-1.0000000000000004,6.661338147750939e-16,-1.85406975049081e-15}},{{-7.771561172376093e-16,-1.0000000000000004,2.020408977628163e-16}},{{-1.7406871441220667e-15,3.5783582556307834e-16,1.0000000000000004}}}},{{{{-1.0000000000000004,2.5954630137242344e-16,2.202611201611221e-15}},{{-1.554312234475219e-15,0.49999999999999994,-0.8660254037844389}},{{-1.2965979342720042e-15,-0.8660254037844392,-0.5000000000000001}}}},{{{{-0.74535599249993,0.6666666666666671,1.591988538067385e-15}},{{0.6666666666666672,0.7453559924999301,-8.881784197001252e-16}},{{-1.6653345369377348e-15,5.551115123125783e-16,-1.0000000000000004}}}}};

void vec3_stereo(struct vec2 *o, const struct vec3 *v) {
  double k = 1 - v->v[2];
  o->v[0] = v->v[0] / k;
  o->v[1] = v->v[1] / k;
}

void moveTo(cairo_t *cr, const struct vec2 *p) {
  cairo_move_to(cr, coordx(p->v[0]), coordy(p->v[1]));
}

void lineTo(cairo_t *cr, const struct vec2 *p) {
  cairo_line_to(cr, coordx(p->v[0]), coordy(p->v[1]));
}

void arc(cairo_t *cr, const struct vec2 *p, double r, double a0, double a1) {
  cairo_arc(cr, coordx(p->v[0]), coordy(p->v[1]), dist(r), a0, a1);
}

void arcNegative(cairo_t *cr, const struct vec2 *p, double r, double a0, double a1) {
  cairo_arc_negative(cr, coordx(p->v[0]), coordy(p->v[1]), dist(r), a0, a1);
}

int segment1(cairo_t *cr, const struct vec2 *u, const struct vec3 *p, const struct vec2 *v, const struct vec3 *q) {
  struct vec3 pq;
  vec3_add(&pq, p, q);
  vec3_normalize(&pq);
  struct vec2 uv, uuv, uvv, uuvv;
  vec3_stereo(&uv, &pq);
  vec2_sub(&uuv, u, &uv);
  vec2_sub(&uvv, v, &uv);
  vec2_sub(&uuvv, u, v);
  double a2 = vec2_normSq(&uuv);
  double b2 = vec2_normSq(&uvv);
  double c2 = vec2_normSq(&uuvv);
  double a = sqrt(a2);
  double b = sqrt(b2);
  double c = sqrt(c2);
  double s = (a + b + c) / 2;
  double d = s * (s - a) * (s - b) * (s - c);
  const struct vec3 inf = {{0,0,1}};
  struct mat33 m;
  vec3_copy(&m.v[0], p);
  vec3_copy(&m.v[1], q);
  vec3_copy(&m.v[2], &inf);
  double k = a * b * c;
  double r = k / (4 * sqrt(d));
  if (d < 1e-12 || r > 1e4 * c || r > 1e5) {
    lineTo(cr, v);
    return 0;
  } else {
    double ex = uuv.v[0];
    double ey = uuv.v[1];
    double e2 = ex * ex + ey * ey;
    double fx = uvv.v[0];
    double fy = uvv.v[1];
    double f2 = fx * fx + fy * fy;
    double ef = 2 * (ex * fy - ey * fx);
    int cw = ef > 0;
    double ox = (fy * e2 - ey * f2) / ef;
    double oy = (ex * f2 - fx * e2) / ef;
    struct vec2 o;
    o.v[0] = ox + uv.v[0];
    o.v[1] = oy + uv.v[1];
    double au = atan2(u->v[1] - o.v[1], u->v[0] - o.v[0]);
    double av = atan2(v->v[1] - o.v[1], v->v[0] - o.v[0]);
    if (cw) {
      arcNegative(cr, &o, r, au, av);
      return -1;
    } else {
      arc(cr, &o, r, au, av);
      return 1;
    }
  }
}

double vec3_dot(const struct vec3 *p, const struct vec3 *q) {
  return p->v[0] * q->v[0] + p->v[1] * q->v[1] + p->v[2] * q->v[2];
}

double vec2_dot(const struct vec2 *p, const struct vec2 *q) {
  return p->v[0] * q->v[0] + p->v[1] * q->v[1];
}

void vec3_slerp(struct vec3 *v, const struct vec3 *p, const struct vec3 *q, double t) {
  double o = acos(vec3_dot(p, q));
  double f = o != 0 ? sin((1 - t) * o) / sin(o) : 1;
  double g = o != 0 ? sin(t * o) / sin(o) : 0;
  for (int i = 0; i < 3; ++i) {
    v->v[i] = f * p->v[i] + g * q->v[i];
  }
  vec3_normalize(v);
}

double segment(cairo_t *cr, const struct vec3 *p, const struct vec3 *q) {
  struct vec2 uu, vv;
  vec3_stereo(&uu, p);
  vec3_stereo(&vv, q);
  return segment1(cr, &uu, p, &vv, q);
}


void vec2_print(FILE *f, const struct vec2 *v) {
  fprintf(f, "( %f , %f )", v->v[0], v->v[1]);
}

void vec3_print(FILE *f, const struct vec3 *v) {
  fprintf(f, "( %f , %f , %f )", v->v[0], v->v[1], v->v[2]);
}

void mat33_print(FILE *f, const struct mat33 *v) {
  vec3_print(f, &v->v[0]);
  fprintf(f, "\t");
  vec3_print(f, &v->v[1]);
  fprintf(f, "\t");
  vec3_print(f, &v->v[2]);
  fprintf(f, "\n");
}

void vec3_swap(struct vec3 *x, struct vec3 *y) {
  double t;
  t = x->v[0]; x->v[0] = y->v[0]; y->v[0] = t;
  t = x->v[1]; x->v[1] = y->v[1]; y->v[1] = t;
  t = x->v[2]; x->v[2] = y->v[2]; y->v[2] = t;
}

void quad(cairo_t *cr, const struct vec3 *w, const struct vec3 *x, const struct vec3 *y, const struct vec3 *z) {
  struct vec2 p;
  vec3_stereo(&p, w);
  moveTo(cr, &p);
  segment(cr, w, x);
  segment(cr, x, y);
  segment(cr, y, z);
  segment(cr, z, w);
  cairo_fill(cr);
}

struct hay {
  // generated values
  double length;
  struct mat33 orientl, orientr;
  struct vec3 center;
  double width;
  // computed values
  double area;
};

void hay_random(struct hay *h) {
  h->length = fmax(pow(double_random(), 2.0) * pi / 3.0, pi/180);
  mat33_random(&h->orientl, &h->orientr, h->length);
  vec3_random(&h->center);
  h->width = 0.00001 * pow(1000.0, pow(double_random(), 2.0));
  h->area = h->length * h->width;
}

int double_compare(double x, double y) {
  if (x <  y) return -1;
  if (x >  y) return  1;
  if (x == y) return  0;
  fprintf(stderr, "oops: double_compare(%f, %f)\n", x, y);
  return 0;
}

int hay_compare(const void *a, const void *b) {
  const struct hay *x = a;
  const struct hay *y = b;
  return double_compare(x->area, y->area);
}

struct needle {
  // generated values
  double x, y, a;
};

void needle_random(struct needle *n) {
  n->x = (double_random() * 0.75 + 0.125) * sizex + clipx;
  n->y = (double_random() * 0.75 + 0.125) * sizey + clipy;
  n->a = 2 * pi * double_random();
}

struct haystack {
  // input values
  int page;
  // generated values
  struct mat33 transforml, transformr;
  struct hay lines[count];
  struct needle ndle;
  // computed values
  double area;
};

void haystack_random(struct haystack *hs, int page) {
  unsigned long int seed = page;
  gsl_rng_set(rng, seed);
  hs->page = page;
  hs->area = 0;
  mat33_random(&hs->transforml, &hs->transformr, 2 * pi);
  for (int i = 0; i < count; ++i) {
    hay_random(&hs->lines[i]);
    hs->area += hs->lines[i].area;
  }
  needle_random(&hs->ndle);
}

void haystack_sort(struct haystack *hs) {
  qsort(&hs->lines[0], count, sizeof(struct hay), hay_compare);
}

void needle_draw(cairo_t *cr, struct needle *n) {
  cairo_save(cr);
  cairo_translate(cr, n->x, n->y);
  cairo_rotate(cr, n->a);
  cairo_scale(cr, 12 * pt, 12 * pt);
  // based on ginger coons openclipart sewing needle (public domain)
  cairo_scale(cr, 4.0, 4.0);
  cairo_translate(cr, -1.0, 0.0);
  cairo_scale(cr, 2.0 / 768.0, 2.0 / 768.0);
  cairo_translate(cr, 460.0, -514.0);
  cairo_rotate(cr, 52.871601612 * pi / 180.0);
#define L(x,y) cairo_rel_line_to(cr,x,y);
  cairo_move_to(cr, 143.64116, 681.89804);
  L(5.67037,-10.40189)
  L(6.66261,-10.95148)
  L(13.33871,-18.97823)
  L(11.43321,-16.00525)
  L(14.58843,-20.48823)
  L(28.73517,-37.9828)
  L(21.60053,-29.05343)
  L(48.28335,-63.48856)
  L(26.29727,-34.54568)
  L(56.3567,-74.23486)
  L(78.80136,-104.08028)
  L(49.1595,-65.08767)
  L(6.77401,-9.88541)
  L(10.69553,-14.13048)
  L(11.23204,-16.06296)
  L(9.08516,-13.57657)
  L(14.78895,-21.8822)
  L(12.14588,-17.03446)
  L(9.67111,-12.77337)
  L(9.6189,-11.137351)
  L(6.80821,-6.337669)
  L(3.25057,-1.970079)
  L(3.56377,0.605429)
  L(2.42316,0.913137)
  L(0.88314,2.739049)
  L(0.79331,3.747456)
  L(-0.769,4.133711)
  L(-2.78581,6.040567)
  L(-5.01123,8.105247)
  L(-8.91683,12.989843)
  L(-11.21185,15.23419)
  L(-12.20427,15.78396)
  L(-11.83956,15.27173)
  L(-10.85983,13.2487)
  L(-8.66415,9.95988)
  L(-5.0119,6.65362)
  L(-15.3388,20.8898)
  L(-45.23872,59.39112)
  L(-59.54815,78.90706)
  L(-42.74299,55.7527)
  L(-49.03289,65.34175)
  L(-48.42685,63.22982)
  L(-45.36525,59.13687)
  L(-40.30112,50.84727)
  L(-18.31997,21.92123)
  L(-7.40954,8.4331)
  L(-6.4107,6.46985)
  L(-4.32795,3.51156)
  L(-0.95472,1.17748)
  L(-0.99236,0.54977)
  L(-0.97604,-0.8974)
  cairo_close_path(cr);
  cairo_rel_move_to(cr, 398.51769,-528.84985);
  L(0.53301,0.98756)
  L(1.87001,0.75436)
  L(1.63676,1.50389)
  L(3.83587,-4.70505)
  L(11.52288,-15.56254)
  L(14.33511,-18.90998)
  L(11.59676,-16.57516)
  L(5.13933,-7.034319)
  L(1.02785,-2.190656)
  L(-1.27204,-2.016101)
  L(-2.03904,-1.619366)
  L(-1.52542,-0.437766)
  L(-2.45865,2.814215)
  L(-0.99169,2.001464)
  L(-3.70437,4.94237)
  L(-6.16787,7.773329)
  L(-6.58549,8.46991)
  L(-5.94591,8.45391)
  L(-6.774,9.88543)
  L(-5.59312,7.92011)
  L(-6.93758,10.45534)
  L(-1.50316,3.08842)
  cairo_close_path(cr);
#undef L
  cairo_restore(cr);
  cairo_set_source_rgba(cr, 1,1,1, 1);
  cairo_set_line_width(cr, 1 * pt);
  cairo_stroke_preserve(cr);
  double g = 0.75;
  cairo_set_source_rgba(cr, g,g,g, 1);
  cairo_fill_preserve(cr);
  cairo_set_line_width(cr, 0.5 * pt);
  cairo_set_source_rgba(cr, 0,0,0, 1);
  cairo_stroke(cr);
}

void hay_draw(cairo_t *cr, struct hay *h, struct mat33 *m0, double f) {
  if (! (0 < f)) {
    return;
  }
  // line endpoints
  struct vec3 p2, p3;
  mat33_multmv(&p2, &h->orientl, &h->center);
  mat33_multmv(&p3, &h->orientr, &h->center);
  vec3_normalize(&p2);
  vec3_normalize(&p3);
  for (int j = 0; j < 60; ++j) {
    // rotate line symmetrically
    struct mat33 m;
    mat33_multmm(&m, m0, &poly_5_3[j]);
    struct vec3 u, v, d, du, dv, x, y, z, w, midpoint;
    mat33_multmv(&u, &m, &p2);
    mat33_multmv(&v, &m, &p3);
    vec3_normalize(&u);
    vec3_normalize(&v);
    // expand to rectangle
    vec3_sub(&d, &u, &v);
    vec3_normalize(&d);
    vec3_cross(&du, &d, &u);
    vec3_normalize(&du);
    vec3_mul(&du, h->width);
    vec3_cross(&dv, &d, &v);
    vec3_normalize(&dv);
    vec3_mul(&dv, h->width);
    vec3_add(&w, &u, &du);
    vec3_normalize(&w);
    vec3_sub(&x, &u, &du);
    vec3_normalize(&x);
    vec3_sub(&y, &v, &dv);
    vec3_normalize(&y);
    vec3_add(&z, &v, &dv);
    vec3_normalize(&z);
    // shrink rectangle
    double len = h->length;
    if (f <= 0.5) {
      vec3_slerp(&y, &x, &y, 2 * f);
      vec3_slerp(&z, &w, &z, 2 * f);
      len *= 2 * f;
    } else if (f <= 1) {
      vec3_slerp(&x, &y, &x, 2 - 2 * f);
      vec3_slerp(&w, &z, &w, 2 - 2 * f);
      len *= 2 - 2 * f;
    }
    // clip/split against a small polar circle
#ifdef VIDEO
    double k = factor * (1 + 81/256) * sqrt(16*16+9*9)/16.0 / 2.0;
#else
    double k = factor * sqrt(5) / 2.0;
#endif
    double cy = (k * k - 1) / (k * k + 1);
    double cx = 2 * k / (k * k + 1);
    double cz = cos(len) * cy - sin(len) * cx;
    struct vec3 dx[2];
    vec3_cross(&dx[0], &x, &y);
    vec3_normalize(&dx[0]);
    vec3_cross(&dx[1], &w, &z);
    vec3_normalize(&dx[1]);
    vec3_add(&midpoint, &w, &x);
    vec3_add(&midpoint, &midpoint, &y);
    vec3_add(&midpoint, &midpoint, &z);
    vec3_normalize(&midpoint);
    if (vec3_normSq(&dx[0]) > 0 && vec3_normSq(&dx[1]) > 0) {
    if (fabs(dx[0].v[2]) <= cx && fabs(dx[1].v[2]) <= cx && midpoint.v[2] >= cz) {
      // both great circles intersect the polar circle // && the segment is close enough that it could
      // find both intersections of each great circle with the polar circle
      struct vec3 u1[2], v1[2];
      int isec = 1;
      for (int m = 0; m < 2; ++m) {
        double a = dx[m].v[0] * dx[m].v[0] + dx[m].v[1] * dx[m].v[1];
        double b = 2 * dx[m].v[1] * dx[m].v[2] * cy;
        double c = dx[m].v[2] * dx[m].v[2] * cy * cy  + dx[m].v[0] * dx[m].v[0] * (cy * cy - 1);
        double d = b * b - 4 * a * c;
        if (d < 0) {
          isec = 0;
          d = 0;
        }
        u1[m].v[2] = cy;
        u1[m].v[1] = (-b - sqrt(d)) / (2 * a);
        u1[m].v[0] = - (dx[m].v[2] * cy + dx[m].v[1] * u1[m].v[1]) / dx[m].v[0];
        vec3_normalize(&u1[m]);
        v1[m].v[2] = cy;
        v1[m].v[1] = (-b + sqrt(d)) / (2 * a);
        v1[m].v[0] = - (dx[m].v[2] * cy + dx[m].v[1] * v1[m].v[1]) / dx[m].v[0];
        vec3_normalize(&v1[m]);
        // find segment order
        struct vec3 l;
        vec3_cross(&l, &u1[m], &v1[m]);
        vec3_normalize(&l);
        if (vec3_dot(&dx[m], &l) <= 0) {
          vec3_swap(&u1[m], &v1[m]);
        }
        // segments are u,u1 and v1,v
      }
      if (isec) {
        // if u1 between u and v, and u outside polar circle, draw u u1
        double ll = vec3_dot(&u, &v);
        if (ll <= vec3_dot(&x, &u1[0]) && ll <= vec3_dot(&y, &u1[0]) && w.v[2] <= cy && x.v[2] <= cy) {
          quad(cr, &w, &x, &u1[0], &u1[1]);
        }
        // if v1 between u and v, and v outside polar circle, draw v v1
        if (ll <= vec3_dot(&x, &v1[0]) && ll <= vec3_dot(&y, &v1[0]) && y.v[2] <= cy && z.v[2] <= cy) {
          quad(cr, &z, &y, &v1[0], &v1[1]);
        }
        // if neither u1, v1 between u v, and u,v outside polar circle, segment didn't intersect polar circle
        if (!(ll <= vec3_dot(&x, &u1[0]) && ll <= vec3_dot(&y, &u1[0])) &&
            !(ll <= vec3_dot(&x, &v1[0]) && ll <= vec3_dot(&y, &v1[0])) &&
             w.v[2] <= cy && x.v[2] <= cy && y.v[2] <= cy && z.v[2] <= cy) {
          quad(cr, &w, &x, &y, &z);
        }
      // both u,v inside circle, don't draw anything...
      } else {
        quad(cr, &w, &x, &y, &z);
      }      
    } else {
      // quad doesn't intersect polar circle, just draw it
      quad(cr, &w, &x, &y, &z);
    }
    }
  }
}

int main (int argc, char **argv) {
  rng = gsl_rng_alloc(gsl_rng_mt19937);
  int page = 0;
  if (argc > 1) {
    page = atoi(argv[1]);
  }

  for (int format = 0; format < 2; ++format) {

    struct haystack hs;
    haystack_random(&hs, page);

    char filename[1024];
    cairo_surface_t *surface = 0;
    cairo_t *cr = 0;
    switch (format) {
      case 0:
        snprintf(filename, 1000, "%d.pdf", page);
        surface = cairo_pdf_surface_create(filename, pagex, pagey);
        cr = cairo_create(surface);
        break;
      case 1:
        snprintf(filename, 1000, "%d.png", page);
        surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24, 144, 204);
        cr = cairo_create(surface);
        cairo_rectangle(cr, -1, -1, 146, 206);
        cairo_set_source_rgba(cr, 1,1,1, 1);
        cairo_fill(cr);
        cairo_scale(cr, 144.0 / pagex, 204.0 / pagey);
        break;
    }

    cairo_rectangle(cr, clipx, clipy, sizex, sizey);
    cairo_clip(cr);
    double f = 0.5;

    cairo_set_fill_rule(cr, CAIRO_FILL_RULE_EVEN_ODD);
    cairo_set_source_rgba(cr, 0,0,0, 1);
    for (int i = 0; i < count; ++i) {
      hay_draw(cr, &hs.lines[i], &hs.transforml, f);
    }
    needle_draw(cr, &hs.ndle);
    // done drawing image
    cairo_reset_clip(cr);

    // add text to page
    cairo_rectangle(cr, clipx, pagey - clipy - pt, sizex, pt);
    cairo_fill(cr);
    const char *tleft = "Haystack Situation";
    char tright[1024];
    snprintf(tright, 1000, "#%d", page);
    cairo_select_font_face(cr, "LMSans10", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);
    cairo_set_font_size(cr, 12 * pt);
    cairo_text_extents_t te;
    cairo_text_extents(cr, tleft, &te);
    cairo_move_to(cr, clipx - te.x_bearing, (pagey + sizey) / 2.0 - te.height / 2.0 - te.y_bearing);
    cairo_show_text(cr, tleft);
    cairo_text_extents(cr, tright, &te);
    cairo_move_to(cr, pagex - clipx - te.width - te.x_bearing, (pagey + sizey) / 2.0 - te.height / 2.0 - te.y_bearing);
    cairo_show_text(cr, tright);

    switch (format) {
      case 0:
        cairo_show_page(cr);
        break;
      case 1:
        cairo_surface_write_to_png(surface, filename);
        break;
    }

    cairo_destroy(cr);
    cairo_surface_destroy(surface);
  }

  gsl_rng_free(rng);
  return 0;
}
